# -*-python-*-
# $Id: s60-locatr.py,v 1.8 2006/09/09 16:54:40 asc Exp $

__package__    = "s60-locatr"
__version__    = "0.1"
__author__     = "Aaron Straup Cope"
__url__        = "http://www.aaronland.info/python/locatr/"
__cvsversion__ = "$Revision: 1.8 $"
__date__       = "$Date: 2006/09/09 16:54:40 $"
__copyright__  = "Copyright (c) 2006 Aaron Straup Cope. Perl Artistic License."

#
# WARNING : SPAGHETTI CODE AHEAD. WELL, MAYBE NOT
# SPAGHETTI BUT NOODLEY. NO REFUNDS ACCEPTED (YET).
#

# TO DO :
#
# * move code into function-related blocks, possibly classes
# * line wrapping 
# * pretty-print lat/long
# * automagic slideshow mode
# * do a sanity check on the error handling/reporting - it's
#   gotten a little out of control
# * redraw screen (title) correctly when requesting next photo
#   from the options menu

import e32
import appuifw
import socket
from key_codes import *

from elementtree.ElementTree import XML
from pdis.xpath import compile

if not e32.in_emulator() :
    import sysinfo
    import graphics
    
import anydbm
import os.path
import httplib
import urllib
import md5
import time

if e32.in_emulator() :
    import pwd
    import posix

#
#
#

class keyboard(object):
    
    def __init__(self,onevent=lambda:None):
        self._keyboard_state={}
        self._downs={}
        self._onevent=onevent

    def handle_event(self,event):
        if event['type'] == appuifw.EEventKeyDown:
            code=event['scancode']
            if not self.is_down(code):
                self._downs[code]=self._downs.get(code,0)+1
            self._keyboard_state[code]=1
        elif event['type'] == appuifw.EEventKeyUp:
            self._keyboard_state[event['scancode']]=0
        self._onevent()

    def is_down(self,scancode):
        return self._keyboard_state.get(scancode,0)

    def pressed(self,scancode):
        if self._downs.get(scancode,0):
            self._downs[scancode]-=1
            return True
        return False

#
#
#

class locatr :

    def __init__ (self) :

        # Flickr
        
        self.__flickr = {'www':'www.flickr.com',
                         'api':'api.flickr.com',
                         'static':'static.flickr.com',
                         'sz':'t'}
        
        self.__apikey = u"787acbb6cee3864e303499ea7a36f5d0"

        # Configs

        self.__approot = os.path.abspath("c")

        if e32.in_emulator() :
            user = pwd.getpwuid(posix.geteuid())
            self.__approot = os.path.abspath(user[5])

        self.__cfg     = None
        self.__cfgpath = u"c:\\locatr.cfg"

        if e32.in_emulator() :
            self.__cfgpath = os.path.abspath(os.path.join(self.__approot, u"locatr.db"))
        
        # BT socket fu
        
        self.__pair = None
        self.__sock = None
        self.__file = None

        # Logging
        
        self.__msg       = u""
        self.__msg_fetch = u"Fetching photo from Flickr..."
        
        self.__msglog  = []
        self.__verbose = False
        
        # Slideshow

        self.__canvas   = None
        self.__img      = None
        self.__photo    = None
        self.__coords   = None
        self.__pause    = False
        self.__kill     = False
        self.__photos   = []
        self.__tmpfiles = []

        self.__fetching = False
        
        # App

        self.lock  = e32.Ao_lock()        
        appuifw.app.exit_key_handler = self.abort
        
        self.__options = []
        
        #

        self.__kb = keyboard()

        self.__xstart = 10
        self.__ystart = 10            
        self.__ypos   = self.__ystart
        self.__yinc   = 15

        if not e32.in_emulator() :
            self.__canvas = appuifw.Canvas(event_callback=self.__kb.handle_event, redraw_callback=self.redraw_handler)
            self.__canvas.clear()

            appuifw.app.body = self.__canvas
            appuifw.app.screen = "large"

            (cx, cy) = self.__canvas.size
            self.__canvas_sz = (cx, cy)
                    
    #
    #
    #
    
    def loop (self) :

        if self.setup() :
            self.run()

        self.lock.wait()

    #
    #
    #
    
    def abort(self) :
        self.__kill = True
        
        self.disconnect_gps_device()
        self.store_cfg()
        self.cleanup()        
        self.lock.signal()

    #
    #
    #

    def cleanup (self, redraw=True) :
        self.__img    = None
        self.__photo  = None
        self.__coords = None
        self.__msglog = []
        self.__ypos   = self.__ystart

        if redraw :
            self.redraw_handler(None)
        
        if len(self.__tmpfiles) == 0 :
            return
        
        self.log("Cleaning up %s temporary files" % len(self.__tmpfiles))

        self.purge_cachefiles(self.__tmpfiles)
        self.log("All done")

    #
    #
    #
    
    def purge_cachefiles (self, files) :

        self.verbose("Clean up %s temp files" % len(files))
        
        while len(files) > 0 :
            upath = unicode(files.pop(0))
            self.verbose("Unlink %s" % upath)
            
            try :
                os.unlink(upath)
            except Exception , e :
                self.verbose("Failed to unlink %s, %s" % (upath, e))
                             
    #
    #
    #
    
    def setup (self) :

        if not self.open_cfg() :
            return False

        if not self.test_gps_connection() :
            return False
        
        if not self.test_http_connection() :
            return False

        #
        
        self.assign_slideshow_menu()
        self.assign_gps_menu()        
        self.assign_verbose_menu((u"Enable verbose mode", self.toggle_verbose), True)            

        #

        (cx, cy) = self.canvas_size()

        if cx > 100 and cy > 100 :
            self.__flickr['sz'] = "m"

        if cx > 240 and cy > 240 :
            big = appuifw.query(u"You have a fancy display. Show big photos?", "query")
                
            if big != None :
                self.__flickr['sz'] = ''
            
        #

        return True

    #
    #
    #

    def test_gps_connection (self) :

        if e32.in_emulator() :
            return True
        
        self.log("Testing GPS connection");

        if not self.connect_gps_device() :
            return False

        return True
    
    #
    #
    #

    def test_http_connection (self) :

        self.log("Testing network connection");
        
        if not e32.in_emulator() :
            if sysinfo.signal() == 0 :
                self.error("No carrier signal")
                return False

        head = self.execute_http_request(self.__flickr['www'], '/', 'HEAD')
        
        if head == False :
            self.error("Unable to connect to flickr.com")
            return False
        
        return True
    
    #
    #
    #

    def open_cfg (self) :
        self.log("Reading configs")

        try :
            self.__cfg = anydbm.open(self.__cfgpath, "c")
        except Exception, e:
            self.error("Failed to open config file, %s" % e)
            return False
        
        return True

    #
    #
    #

    def store_cfg (self) :
        if self.__cfg != None :
            self.__cfg.close()
        
    #
    #
    #
    
    def run (self) :            
        self.log("Ready to show geotagged photos from Flickr!")
        self.log("To begin, select 'Show me photos!' from the Options menu.")
        
    #
    #
    #
    
    def slideshow (self, remove_photos_menu=False) :

        if remove_photos_menu :
            self.remove_photo_menus()
            
        self.cleanup(remove_photos_menu)
        
        (lat, lon) = self.poll()

        if lat == None :
            self.error("Ack! I can't figure out where I am")
            return False
        
        photos = self.fetch_images(lat, lon)

        if len(photos) == 0 :
            self.error("No photos found for this spot : %s, %s" % (lat, lon))
            return False
        
        self.__photos = photos

        self.log("Showing %s Flickr photos around %s, %s" % (len(photos), lat, lon))
        
        self.show_one_photo(True)        
        return True

    #
    #
    #

    def open_photo_in_browser (self) :
        url = self.__photo['link']
        mob = appuifw.query(u"Open in Flickr mobile site", "query")

        if mob != None :
            url = self.__photo['moblink']
            
        self.open_url(url)

    #
    #
    #
    
    def open_url (self, url) :
        if e32.in_emulator() :
            self.log("Open URL %s" % url)
            return
        
        apprun  = 'z:\\system\\programs\\apprun.exe'
        browser = 'z:\\System\\Apps\\Browser\\Browser.app'
        e32.start_exe(apprun, browser + ' "%s"' % url , 1)

    #
    #
    #
    
    def show_one_photo (self, assign_menu=False) :

        if len(self.__photos) == 0 :

            if not appuifw.query(u"Collect more photos?", "query") :
                self.remove_photo_menus()
                return False

            return self.slideshow(True)

        #
        
        if assign_menu :
            self.assign_photo_menus()

        #

        if self.__fetching :
            self.error("Slow down! I am already getting the next photo.")
            return False

        #
        
        self.__fetching = True
        
        if self.__img != None :
            self.__photo['title'] = self.__msg_fetch
            self.redraw_handler(None, True)
        else :
            self.log(self.__msg_fetch)
            
        #

        try :
            ph  = self.__photos.pop()
            url = ph['url']
        except Exception, e :
            self.error("OMGWTF, %s" % e)
            self.__fetching = False            
            return False
        
        data = self.retrieve_image(url)

        if data == None :
            self.__fetching = False
            return False

        store = self.storage_path(url)

        if not self.store_image(store, data) :
            self.__fetching = False            
            return False

        img = self.load_image(store)

        if img == None :
            self.error("Failed to load %s" % store)
            self.__fetching = False            
            return False

        if e32.in_emulator() :
            self.log("render %s" % store)
        else :
            self.render_image(img)

        self.__fetching = False
            
        self.set_title(ph['title'])
        
        #
        
        uid = self.digesterize(ph['id'])
        
        self.__cfg[uid] = unicode(time.time())
        self.__photo    = ph

        self.watch_keyboard_events()            
        return True

    #
    #
    #
    
    def set_title (self, title) :

        if self.__fetching and title != self.__msg_fetch :
            return
        
        if e32.in_emulator() :
            self.log("title : %s" % title)
            return

        (cx, cy) = self.canvas_size(20)
        
        self.__canvas.text((self.__xstart, cy), unicode(title))
        
    #
    #
    #
    
    def watch_keyboard_events (self) :

        if self.__kb.pressed(EScancodeUpArrow) :
            return self.open_photo_in_browser()

        if self.__kb.pressed(EScancodeLeftArrow) :
            return self.slideshow(True)

        if self.__kb.pressed(EScancodeRightArrow) :
            return self.show_one_photo()

        e32.ao_sleep(1, self.watch_keyboard_events)
        
    #
    #
    #

    def connect_gps_device (self) :

        if not self.find_gps_device() :
            return False

        if not self.connect_gps_device_bt() :
            return False

        return True

    #
    #
    #
    
    def disconnect_gps_device (self) :

        if self.__sock == None :
            return True

        try :
            self.__sock.close()
        except Exception, e :
            self.error("Failed to disconnect from GPS device, %s" % e)
            return False

        self.__file = None
        self.__sock = None
        return True
    
    #
    #
    #

    def fetch_images (self, lat, lon, acc=16) :
        photos = []
        
        boundingbox = self.mk_bounding_box(lat, lon)

        if boundingbox == None :
            return photos
        
        self.verbose("bbox : %s" % boundingbox)

        cur_page  = 1
        num_pages = None
        
        while len(photos) < 10 :

            rem = 10 - len(photos)
            msg = ""

            if cur_page == 1 :
                self.log("Retrieving 10 photos from Flickr.")
            else :
                self.log("Some of those have been viewed already.")
                self.log("Retrieving another %s photos from Flickr." % rem)
            
            xml = self.api_call('flickr.photos.search',
                                bbox=boundingbox,
                                accuracy=acc,
                                page=cur_page,
                                per_page=10,
                                min_upload_date=892800,
                                extras='geo')

            #
            
            if num_pages == None :
                xp = compile("/rsp/photos")
                res = xp.evaluate(xml)
                num_pages = res[0].attrib['pages']

            #
            
            xp  = compile("/rsp/photos/photo")
            res = xp.evaluate(xml)
        
            for ph in res :

                url = "/%s/%s_%s_%s.jpg" % (ph.attrib['server'] , ph.attrib['id'], ph.attrib['secret'], self.__flickr['sz'])
                
                if self.__flickr['sz'] == '' :
                    url = "/%s/%s_%s.jpg" % (ph.attrib['server'] , ph.attrib['id'], ph.attrib['secret'])                    

                id  = ph.attrib['id']
                uid = self.digesterize(id)
                
                if self.__cfg.has_key(uid) :
                    self.verbose("already seen %s" % url)
                    continue
                
                lat   = ph.attrib['latitude']
                lon   = ph.attrib['longitude']
                title = ph.attrib['title']
                
                link    = "http://%s/photos/%s/%s" % (self.__flickr['www'], ph.attrib['owner'], ph.attrib['id'])
                moblink = "http://%s/mob/photo.gne?id=%s" % (self.__flickr['www'], ph.attrib['id'])
                
                photos.append({'url':url,'link':link,'moblink':moblink, 'lat':lat,'lon':lon,'title':title,'id':id})
                    
                if len(photos) == 10 :
                    break

            #
            
            cur_page += 1

            self.verbose("pages : %s / %s" % (cur_page, num_pages))
            
            if cur_page > num_pages :
                break

        #
        #
        #

        return photos
    
    #
    #
    #
    
    def poll (self) :

        # return (37.750917, -122.421823)
        
        if e32.in_emulator() :
            return (37.750917, -122.421823)
        
        if self.__sock == None :
            if not self.connect_gps_device() :
                self.error("Unable to connect to GPS, giving up")
                return (None, None)

        #

        data = self.readlines("$GPRMC")

        lat = None
        lon = None
        
        if data != "" :
            self.verbose(data)

            try :
                (lat, lon) = self.parse_gps_data(data)
                self.verbose("%s, %s" % (lat, lon))
                
            except Exception, e :
                self.error("parse error, %s" % e)
            
        #
        
        self.disconnect_gps_device()
        return (lat, lon)

    #
    #
    #
    
    def mk_bounding_box (self, lat, lon, sz=2) :

        edge   = float(sz) / float(2)
        offset = float(edge) / float(69)

        try :
            lat = float(lat)
            lon = float(lon)
        except Exception, e :
            self.error("Failed to format lat, long values : %s" % e)
            return None
        
        swlat = float("%.6f" % (lat - offset))
        swlon = float("%.6f" % (lon - offset))
        nelat = float("%.6f" % (lat + offset))
        nelon = float("%.6f" % (lon + offset))

        return "%s,%s,%s,%s" % (swlon, swlat, nelon, nelat)
        
    #
    #
    #
    
    def parse_gps_data (self, gps_data) :

        # Thank you :
        # http://positio.rista.net/en/pys60gps/
        
        # Expect 11 ','-delimiters (NMEA 2.2)
        # Expect 12 ','-delimiters (NMEA 2.3)
        
        if (gps_data.count(",") == 11) :
            (GPRMC,hhmmssss,status,lat,l1,lon,l2,speed_knots,course,ddmmyy,mvd,mdr_cs)=gps_data.split(",")
        elif (gps_data.count(",") == 12) :
            (GPRMC,hhmmssss,status,lat,l1,lon,l2,speed_knots,course,ddmmyy,mvd,mdr,fix_cs)=gps_data.split(",")
        else:
            self.error("Invalid GPRMC string")

        if (status is 'A' and lat is not "" and lon is not ""):
            lat, lon = self.parse_gps_latlong(lat, l1, lon, l2)

        return (lat, lon)

    #
    #
    #

    def parse_gps_latlong (self, lat, l1, lon, l2):
        
        try:
            latitude = float(lat[0:2]) + (float(lat[2:4] + "." + lat[5:9])/60)
            longitude = float(lon[0:3]) + (float(lon[3:5] + "." + lon[6:10])/60)

            if l1 is "S":
                latitude = -latitude

            if l2 is "W":
                longitude = -longitude

            latitude  = float("%.6f" % latitude)
            longitude = float("%.6f" % longitude)
            
            return latitude, longitude
        except:
            return (None, None)
        
    #
    #
    #
    
    def readlines (self, type, iterations=50) :
        self.verbose("Whispering sweet nothings...")
        
        count = 0
        data  = ""
        
        while count < iterations :

            try :
                ln = self.__file.readline()
                ln.strip()
                
                if ln.startswith(type) :
                    data = ln
                    break

            except socket.error, e :
                self.error("There was an error talk to the GPS device : %s" % e)
                break
            except Excpetion, e :
                self.error("Hrm - I have no idea what's going on : %s" % e)
                break
            
            count += 1

        if data == "" :
            self.error("Failed to receive any suitable data after %s iterations" % iterations)
            
        return data
    
    #
    #
    #
        
    def find_gps_device (self) :

        self.verbose("finding GPS device")
        
        if self.__cfg.has_key('gps_address') :
            self.verbose("Using default GPS device")
            
            self.__pair = (self.__cfg['gps_address'], int(self.__cfg['gps_service']))
            return True

        return self.pair_gps_device()
        
    #
    #
    #
    
    def pair_gps_device (self) :
        self.verbose("Pairing with unknown GPS device")
        
        try :
            address, services=socket.bt_discover()
        except Exception, e :
            self.error("Failed to discover GPS device, %s" % e)
            return False

        service = self.choose_bt_service(services)

        if service == None :
            self.error("Failed to discover GPS service")
            return False

        save = appuifw.query(u"Store as default GPS device?", "query")

        if save != None :
            self.__cfg['gps_address'] = unicode(address)
            self.__cfg['gps_service'] = unicode(service)

            self.log("store address %s" % address)
            self.log("store serivce %s" % service)

        self.__pair = (address, service)            

        self.verbose("Paired with %s" % address)
        return True

    #
    #
    #
    
    def choose_bt_service (self, services) :

        if len(services)>1:
            choices=services.keys()
            choices.sort()

            choice=appuifw.popup_menu([unicode(services[x])+": "+x for x in choices], u'Choose GPS Service:')

            if choice == None :
                return None
            
            return services[choices[choice]]
        
        else:
            return services.values()[0]
        
    #
    #
    #

    def connect_gps_device_bt (self) :

        self.verbose("Connecting to GPS device")
        
        try :
            sock=socket.socket(socket.AF_BT,socket.SOCK_STREAM)
        except Exception, e :
            self.error("Failed to create socket, %s" % e)
            return False

        try : 
            sock.connect(self.__pair)
        except Exception, e :
            self.error("Failed to connect to GPS device, %s" % e)
            return False

        self.__sock = sock
        self.__file = self.__sock.makefile(512)
        
        return True

    #
    #
    #

    def api_call(self, method, **kwargs) :

        kwargs['method']  = method
        kwargs['api_key'] = self.__apikey
        
        url = "/services/rest/?%s" % urllib.urlencode(kwargs)

        self.verbose(url)

        res = self.execute_http_request(self.__flickr['api'], url)

        if res == None :
            return None
        
        raw = res.read()
        # self.verbose(raw)
        
        try :
            xml = XML(raw)
        except Exception, e :
            self.log("XML parse error : %s" % e)
            return None

        if self.is_api_error(xml) :
            return None

        return xml

    #
    #
    #
    
    def is_api_error (self, xml) :

        xp  = compile("/rsp")
        rsp = xp.evaluate(xml)[0]

        if rsp.attrib['stat'] == 'fail' :
            err = rsp.find('err')
            self.log("API error : %s (%s)" % (err.attrib['code'], err.attrib['msg']))
            return True

        return False

    #
    #
    #
    
    def execute_http_request (self, endpoint, url, method='GET') :

        try :
            conn = httplib.HTTPConnection(endpoint)
            conn.request(method, url)
        except Exception, e :
            self.log("HTTP request failed : %s" % e)
            return None
            
        return conn.getresponse()
    
    #
    #
    #

    def toggle_verbose (self) :

        if self.__verbose :
            self.__verbose = False
            self.assign_verbose_menu((u"Enable verbose mode", self.toggle_verbose))            
        else :
            self.__verbose = True
            self.assign_verbose_menu((u"Disable verbose mode", self.toggle_verbose))

    #
    #
    #

    def assign_slideshow_menu (self, startup=False) :
        self.__options.insert(0, (u"Show me photos!", self.slideshow))
        self.assign_menu()

    #
    #
    #

    def assign_gps_menu (self) :
        if not e32.in_emulator() :
            self.__options.append((u'Pair with GPS device', self.pair_gps_device))
            self.assign_menu()

    #
    #
    #
    
    def assign_photo_menus (self) :
        self.__options.insert(1, (u"Next photo", self.show_one_photo))
        self.__options.insert(2, (u"Open photo", self.open_photo_in_browser))        
        self.assign_menu()

    #
    #
    #
    
    def remove_photo_menus (self) :
        self.__options = self.__options[3:]
        self.assign_slideshow_menu()
            
    #
    #
    #
    
    def assign_verbose_menu (self, menu_item, startup=False) :

        if startup :
            self.__options.append(menu_item)
        else :
            idx = len(self.__options) - 1
            self.__options[idx] = menu_item

        self.assign_menu()

    #
    #
    #

    def assign_menu (self) :
        appuifw.app.menu = self.__options
        
    #
    #
    #
    
    def verbose (self, msg) :

        if not self.__verbose :
            return

        self.log(msg)

    #
    #
    #

    def error (self, msg) :

        if e32.in_emulator() :
            self.log(msg)
            return

        appuifw.note(unicode(msg), "error")        
        
    #
    #
    #
    
    def log (self, msg) :

        if e32.in_emulator() :
            self.__msg += unicode("%s\n" % msg)
            self.write(self.__msg)
            return

        ypos = self.__ypos + self.__yinc
        (cx, cy) = self.canvas_size(0)

        if (ypos + 10) >= cy :
            self.canvas_clear()
            self.__msglog = []
            ypos = self.__ystart

        self.__ypos = ypos
        umsg = unicode(msg)
        
        self.__canvas.text((self.__xstart, ypos), umsg)
        self.__msglog.append((ypos, umsg))

    #
    #
    #
    
    def relog (self) :
        self.canvas_clear()
        
        for ypos, umsg in self.__msglog :
            self.__canvas.text((self.__xstart, ypos), umsg)
            
    #
    #
    #
    
    def write(self, txt) :
        txt = unicode(txt)
        
        if e32.in_emulator() :
            t = appuifw.Text()
            t.write(txt)
            appuifw.app.body = t

        else :
            appuifw.app.body = appuifw.Text(txt)

    #
    #
    #

    def storage_path (self, url) :
        hex   = self.digesterize(url)
        fname = u"fl-%s.jpg" % hex

        store = u"c:\\%s" % fname

        if e32.in_emulator() :
            store = os.path.abspath(os.path.join(self.__approot, fname))
            
        return store

    #
    #
    #

    def digesterize (self, url) :
        return md5.md5(url).hexdigest()
    
    #
    #
    #

    def retrieve_image (self, path) :
        self.verbose("retrieve %s" % path)
        
        data = self.execute_http_request(self.__flickr['static'], path)
        return data

    #
    #
    #

    def load_image (self, path) :

        if not self.carry_on() :
            return None

        self.verbose("load image, %s" % path)
        
        if e32.in_emulator() :
            return True
        
        try :
            image = self.open_image(path)
        except Exception, e :
            self.error("Failed to open image %s, %s" % (store, e))
            return None

        try :
            image = self.resize_image(image)
        except Exception, e :
            self.error("Failed to resize image, %s" % e)
            return None
            
        return image

    #
    #
    #

    def store_image (self, path, data) :

        if not self.carry_on() :
            return False

        self.verbose("store %s" % path)
        
        try :
            fh = open(path, "w")
            fh.write(data.read())
            fh.close()
        except Exception, e :
            self.error("Failed to store image locally in %s, %s" % (path, e))
            return False

        if path not in self.__tmpfiles :
            self.__tmpfiles.append(path)

        return True

    #
    #
    #
    
    def open_image (self, path) :

        if not self.carry_on() :
            return None

        self.verbose("open %s" % path)
        return graphics.Image.open(path)

    #
    #
    #
    
    def resize_image (self, image) :

        if not self.carry_on() :
            return None

        self.verbose("resize image")
        
        (ix, iy) = image.size
        (cx, cy) = self.canvas_size()
        
        if ix <= cx and iy <= cy :
            return image

        if ix > iy :
            rx = int(cx)
            ry = int(cy * (float(cx) / float(ix)))
        else :
            rx = int(cx * (float(cy) / float(iy)))
            ry = int(cy)

        return image.resize((rx, ry), callback=None, keepaspect=1)

    #
    #
    #

    def centering_coords (self, image) :

        (ix, iy) = image.size
        (cx, cy) = self.canvas_size()
        
        if ix == cx and iy == cy :
            return (0, 0)
        
        rx = int(float(cx - ix) / 2)
        ry = int(float(cy - iy) / 2)        

        return (rx, ry)
    
    #
    #
    #

    def render_image (self, image) :

        self.verbose("Render image")
        
        if not self.carry_on() :
            return False

        coords = self.centering_coords(image)
        
        self.canvas_clear()
        self.blit(image, coords)
        
    #
    #
    #
    
    def blit(self, image, coords) :
        self.__img    = image
        self.__coords = coords

        self.__canvas.blit(image, target=coords)
        
    #
    #
    #
        
    def redraw_handler (self, rect, clear=False) :

        if not self.carry_on() :
            return None

        if self.__img != None :

            if clear :
                self.canvas_clear()
                
            self.blit(self.__img, self.__coords)
            self.set_title(self.__photo['title'])
        else :
            self.relog()
            
    #
    #
    #

    def canvas_clear (self) :
        if e32.in_emulator() :
            return

        self.__canvas.clear()

    #
    #
    #
    
    def canvas_size (self, adjusted=30) :

        if e32.in_emulator() :
            return (240, 180)
        
        (cx, cy) = self.__canvas_sz

        if adjusted > 0 :
            cy -= adjusted

        return (cx, cy)
    
    #
    #
    #
    
    def carry_on (self) :
        if self.__kill :
            return False
        
        return True

    #
    #
    #
    
if __name__ == "__main__" :
    app = locatr()
    app.loop()
